import json
from collections.abc import Iterator
from functools import cache
from http import HTTPStatus
from typing import Any, Type, TypeVar

import httpx
from pydantic import AnyUrl, BaseModel, ValidationError

from enlyze.auth import TokenAuth
from enlyze.constants import ENLYZE_BASE_URL, HTTPX_TIMEOUT, TIMESERIES_API_SUB_PATH
from enlyze.errors import EnlyzeError, InvalidTokenError
from enlyze.timeseries_api.models import TimeseriesApiModel

T = TypeVar("T", bound=TimeseriesApiModel)


class _PaginatedResponse(BaseModel):
    next: AnyUrl | None
    data: list[Any] | dict[str, Any]


class TimeseriesApiClient:
    """Client class encapsulating all interaction with the Timeseries API

    :param token: API token for the ENLYZE platform
    :param base_url: Base URL of the ENLYZE platform
    :param timeout: Global timeout for all HTTP requests sent to the Timeseries API

    """

    def __init__(
        self,
        token: str,
        *,
        base_url: str | httpx.URL = ENLYZE_BASE_URL,
        timeout: float = HTTPX_TIMEOUT,
    ):
        self._client = httpx.Client(
            auth=TokenAuth(token),
            base_url=httpx.URL(base_url).join(TIMESERIES_API_SUB_PATH),
            timeout=timeout,
        )

    @cache
    def _full_url(self, api_path: str) -> str:
        """Construct full URL from relative URL"""
        return str(self._client.build_request("", api_path).url)

    def get(self, api_path: str, **kwargs: Any) -> Any:
        """Wraps :meth:`httpx.Client.get` with defensive error handling

        :param api_path: Relative URL path inside the Timeseries API (or a full URL)

        :raises: :exc:`~enlyze.errors.EnlyzeError` on request failure

        :raises: :exc:`~enlyze.errors.EnlyzeError` on non-2xx status code

        :raises: :exc:`~enlyze.errors.EnlyzeError` on non-JSON payload

        :returns: JSON payload of the response as Python object

        """

        try:
            response = self._client.get(api_path, **kwargs)
        except Exception as e:
            raise EnlyzeError(
                "Couldn't read from the Timeseries API "
                f"(GET {self._full_url(api_path)})",
            ) from e

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code in (
                HTTPStatus.UNAUTHORIZED,
                HTTPStatus.FORBIDDEN,
            ):
                raise InvalidTokenError
            else:
                raise EnlyzeError(
                    f"Timeseries API returned error {response.status_code}"
                    f" (GET {self._full_url(api_path)})"
                ) from e

        try:
            return response.json()
        except json.JSONDecodeError as e:
            raise EnlyzeError(
                "Timeseries API didn't return a valid JSON object "
                f"(GET {self._full_url(api_path)})",
            ) from e

    def get_paginated(
        self, api_path: str, model: Type[T], **kwargs: Any
    ) -> Iterator[T]:
        """Retrieve objects from paginated Timeseries API endpoint via HTTP GET

        :param api_path: Relative URL path inside the Timeseries API
        :param model: Class derived from
            :class:`~enlyze.timeseries_api.models.TimeseriesApiModel`

        :raises: :exc:`~enlyze.errors.EnlyzeError` on invalid pagination schema

        :raises: :exc:`~enlyze.errors.EnlyzeError` on invalid data schema

        :raises: see :py:meth:`get` for more errors raised by this method

        :returns: Instances of ``model`` retrieved from the ``api_path`` endpoint

        """

        url = api_path

        while True:
            response_body = self.get(url, **kwargs)

            try:
                paginated_response = _PaginatedResponse.parse_obj(response_body)
            except ValidationError as e:
                raise EnlyzeError(
                    f"Paginated response expected (GET {self._full_url(url)})"
                ) from e

            page_data = paginated_response.data
            if not page_data:
                break

            # if `data` is a list we assume there are multiple objects inside.
            # if `data` is a dict then we treat it as only one object
            page_data = page_data if isinstance(page_data, list) else [page_data]

            for elem in page_data:
                try:
                    yield model.parse_obj(elem)
                except ValidationError as e:
                    raise EnlyzeError(
                        f"Timeseries API returned an unparsable {model.__name__} "
                        f"object (GET {self._full_url(api_path)})"
                    ) from e

            if not paginated_response.next:
                break

            url = str(paginated_response.next)



import logging
from datetime import datetime, timezone
from functools import cache
from typing import Iterator, Optional
from uuid import UUID

import enlyze.models as user_models
import enlyze.timeseries_api.models as api_models
from enlyze.errors import EnlyzeError
from enlyze.timeseries_api.client import TimeseriesApiClient


def _ensure_datetime_aware(dt: datetime) -> datetime:
    """Make the returned datetime timezone aware.

    Naive datetime will be assumed to be in local timezone and then converted to aware
    datetime expressed in UTC.

    """

    return dt.astimezone(timezone.utc)


class EnlyzeClient:
    """Main entrypoint for interacting with the ENLYZE platform.

    You should instantiate it only once and use it for all requests to make the best use
    of connection pooling. This client is thread-safe.

    :param token: API token for the ENLYZE platform

    """

    def __init__(self, token: str) -> None:
        self._client = TimeseriesApiClient(token=token)

    def _get_sites(self) -> Iterator[api_models.Site]:
        """Get all sites from the API"""
        return self._client.get_paginated("sites", api_models.Site)

    @cache
    def get_sites(self) -> list[user_models.Site]:
        """Retrieve all :ref:`sites <site>` of your organization.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Sites of your organization
        :rtype: list[:class:`~enlyze.models.Site`]

        """
        return [site.to_user_model() for site in self._get_sites()]

    def _get_appliances(self) -> Iterator[api_models.Appliance]:
        """Get all appliances from the API"""
        return self._client.get_paginated("appliances", api_models.Appliance)

    @cache
    def get_appliances(
        self, site: Optional[user_models.Site] = None
    ) -> list[user_models.Appliance]:
        """Retrieve all :ref:`appliances <appliance>`, optionally filtered by site.

        :param site: Only get appliances of this site. Gets all appliances of the
            organization if None.
        :type site: :class:`~enlyze.models.Site` or None

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Appliances
        :rtype: list[:class:`~enlyze.models.Appliance`]

        """

        if site:
            sites_by_id = {site._id: site}
        else:
            sites_by_id = {site._id: site for site in self.get_sites()}

        appliances = []
        for appliance_api in self._get_appliances():
            site_ = sites_by_id.get(appliance_api.site)
            if not site_:
                continue

            appliances.append(appliance_api.to_user_model(site_))

        return appliances

    def _get_variables(self, appliance_uuid: UUID) -> Iterator[api_models.Variable]:
        """Get variables for an appliance from the API."""
        return self._client.get_paginated(
            "variables", api_models.Variable, params={"appliance": str(appliance_uuid)}
        )

    def get_variables(
        self, appliance: user_models.Appliance
    ) -> list[user_models.Variable]:
        """Retrieve all variables of an :ref:`appliance <appliance>`.

        :param appliance: The appliance for which to get all variables.
        :type appliance: :class:`~enlyze.models.Appliance`

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Variables of ``appliance``
        :rtype: list[:class:`~enlyze.models.Variable`]

        """
        return [
            variable.to_user_model(appliance)
            for variable in self._get_variables(appliance.uuid)
        ]

    def get_timeseries(
        self, start: datetime, end: datetime, variables: list[user_models.Variable]
    ) -> Optional[user_models.TimeseriesData]:
        """Get timeseries data of :ref:`variables <variable>` for a given time frame.

        Timeseries data for multiple variables can be requested at once. However, all
        variables must belong to the same appliance.

        You should always pass :ref:`timezone-aware datetime
        <python:datetime-naive-aware>` objects to this method! If you don't, naive
        datetime objects will be assumed to be expressed in the local timezone of the
        system where the code is run.

        :param start: Beginning of the time frame for which to fetch timeseries data.
            Must not be before ``end``.
        :param end: End of the time frame for which to fetch timeseries data.
        :param variables: The variables for which to fetch timeseries data.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Timeseries data or ``None`` if API returned no data for the request

        """

        if start.utcoffset() is None or end.utcoffset() is None:
            logging.warning(
                "Passing naive datetime is discouraged, assuming local timezone."
            )

        start = _ensure_datetime_aware(start)
        end = _ensure_datetime_aware(end)

        if start > end:
            raise EnlyzeError("Start must be earlier than end")

        appliance_uuids = frozenset(v.appliance.uuid for v in variables)

        if not appliance_uuids:
            raise EnlyzeError("Need to request at least one variable")

        if len(appliance_uuids) != 1:
            raise EnlyzeError(
                "Cannot request timeseries data for more than one appliance per request"
            )

        pages = self._client.get_paginated(
            "timeseries",
            api_models.TimeseriesData,
            params={
                "appliance": str(next(iter(appliance_uuids))),
                "start_datetime": start.isoformat(),
                "end_datetime": end.isoformat(),
                "variables": ",".join(str(v.uuid) for v in variables),
            },
        )

        try:
            timeseries_data = next(pages)
        except StopIteration:
            return None

        if not timeseries_data.columns:
            return None

        if "time" not in timeseries_data.columns:
            raise EnlyzeError("Timeseries API didn't return timestamps")

        for page in pages:
            timeseries_data.extend(page)

        return timeseries_data.to_user_model(
            start=start,
            end=end,
            variables=variables,
        )



#: Base URL of the ENLYZE platform.
ENLYZE_BASE_URL = "https://app.enlyze.com"

#: URL sub-path where the Timeseries API is deployed on the ENLYZE platform.
TIMESERIES_API_SUB_PATH = "timeseries-service/v1/"

#: HTTP timeout for requests to the Timeseries API.
#:
#: Reference: https://www.python-httpx.org/advanced/#timeout-configuration
HTTPX_TIMEOUT = 30.0
